﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using ELTE.TravelAgency.Admin.Persistence;
using ELTE.TravelAgency.Data;
using ELTE.TravelAgency.Model;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Identity;
using Microsoft.AspNetCore.Mvc.Testing;
using Microsoft.AspNetCore.TestHost;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.DependencyInjection.Extensions;
using Xunit;

namespace ELTE.TravelAgency.Service.Test
{
	public class TravelAgencyIntegrationTest : IDisposable
	{
	    public static IList<City> CityData = new List<City>
	    {
	        new City { Id = 1, Name = "TESTCITY" }
	    };

	    public static IList<Building> BuildingData = new List<Building>
	    {
			new Building
            {
                Id = 1,
                CityId = CityData[0].Id,
                City = CityData[0],
                Name = "TESTBUILDING1",
                SeaDistance = 1,
                Shore = ShoreType.Rocky,
                Features = Feature.CoastService | Feature.MainRoad,
                Comment = "Test Building 1"
            },
            new Building
            {
                Id = 2,
                CityId = CityData[0].Id,
                City = CityData[0],
                Name = "TESTBUILDING2",
                SeaDistance = 10,
                Shore = ShoreType.Gravelly,
                Features = Feature.None,
                Comment = "Test Building 2"
            },
            new Building
            {
                Id = 3,
                CityId = CityData[0].Id,
                City = CityData[0],
                Name = "TESTBUILDING3",
                SeaDistance = 100,
                Shore = ShoreType.Sandy,
                Features = Feature.PrivateParking,
                Comment = "Test Building 3"
            }
		};

		private readonly List<CityDTO> _cityDTOs;
		private readonly List<BuildingDTO> _buildingDTOs;
	    private readonly ITravelAgencyPersistence _persistence;
        private readonly WebApplicationFactory<Program> _server;
        private readonly HttpClient _client;

        public TravelAgencyIntegrationTest()
        {
            _cityDTOs = CityData.Select(city => new CityDTO
            {
                Id = city.Id,
                Name = city.Name
            }).ToList();

            _buildingDTOs = BuildingData.Select(building => new BuildingDTO
            {
                Id = building.Id,
                Name = building.Name,
                City = _cityDTOs.Single(city => city.Id == building.CityId),
                Comment = building.Comment,
                Images = new List<ImageDTO>(),
                LocationX = building.LocationX,
                LocationY = building.LocationY,
                SeaDistance = building.SeaDistance,
                Shore = building.Shore,
                Features = FeatureDTO.Convert(building.Features)
            }).ToList();

            // szerver konfiguráció összeállítása és elindítása
            _server = new WebApplicationFactory<Program>()
                .WithWebHostBuilder(builder =>
                {
                    // teszt szerver konfiguráció eltérései
                    builder.ConfigureTestServices(services =>
                    {
                        // Program osztályban megadott adatbázis kontextus eltávolítása
                        services.RemoveAll(typeof(DbContextOptions<TravelAgencyContext>));
                        // In-memory adatbázis kontextus hozzáadása
                        services.AddDbContext<TravelAgencyContext>(options =>
                            options.UseInMemoryDatabase("TravelAgencyIntegrationTest"));

                        var sp = services.BuildServiceProvider();
                        using (var serviceScope = sp.CreateScope())
                        {
                            // Adatbázis inicializálása
                            var dbContext = serviceScope.ServiceProvider.GetRequiredService<TravelAgencyContext>();
                            dbContext.Database.EnsureCreated();
                            dbContext.Cities.AddRange(CityData);
                            dbContext.Buildings.AddRange(BuildingData);
                            dbContext.SaveChanges();

                            var adminUser = new Guest
                            {
                                UserName = "tester",
                                Name = "Teszt Elek",
                                Email = "teszt@example.com",
                                PhoneNumber = "+36123456789",
                                Address = "Próba utca 42."
                            };
                            var adminPassword = "Teszt123";
                            var adminRole = new IdentityRole<int>("administrator");

                            var userManager = serviceScope.ServiceProvider.GetRequiredService<UserManager<Guest>>();
                            var roleManager = serviceScope.ServiceProvider.GetRequiredService<RoleManager<IdentityRole<int>>>();
                            userManager.CreateAsync(adminUser, adminPassword).Wait();
                            roleManager.CreateAsync(adminRole).Wait();
                            userManager.AddToRoleAsync(adminUser, adminRole.Name).Wait();
                        }
                    });
                    builder.UseEnvironment("Testing");
                });
            
            _client = _server.CreateClient(); // kliens példányosítása
            _client.DefaultRequestHeaders.Accept.Clear();
            _client.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));

            _persistence = new TravelAgencyServicePersistence(_client); // kliens oldali perzisztencia réteg példányosítása
        }

        public void Dispose()
        {
            using (var serviceScope = _server.Services.CreateScope())
            {
                var dbContext = serviceScope.ServiceProvider.GetRequiredService<TravelAgencyContext>();
                dbContext.Database.EnsureDeleted();
            }
        }

        [Fact]
		public async void GetCityTest()
        {
            IEnumerable<CityDTO> result = await _persistence.ReadCitiesAsync();

			// Assert
			Assert.Equal(_cityDTOs, result);
		}

		[Fact]
		public async void GetBuildingTest()
        {
            IEnumerable<BuildingDTO> result = await _persistence.ReadBuildingsAsync();

            // Assert
			Assert.Equal(_buildingDTOs, result);
		}

        [Fact]
        public async void CreateBuildingTestUnauthorized()
        {
            var newBuilding = new BuildingDTO
            {
                City = _cityDTOs[0],
                Name = "TESTBUILDING4",
                SeaDistance = 1000,
                Shore = ShoreType.Rocky,
                Features = FeatureDTO.Convert(Feature.Garden | Feature.SwimmingPool),
                Comment = "Test Building 4"
            };

            bool result = await _persistence.CreateBuildingAsync(newBuilding); ;

            // Assert
            Assert.False(result);
        }

        [Fact]
        public async void CreateBuildingTestAuthorized()
        {
            var newBuilding = new BuildingDTO
            {
                City = _cityDTOs[0],
                Name = "TESTBUILDING4",
                SeaDistance = 1000,
                Shore = ShoreType.Rocky,
                Features = FeatureDTO.Convert(Feature.Garden | Feature.SwimmingPool),
                Comment = "Test Building 4"
            };

            await _persistence.LoginAsync("tester", "Teszt123");
            bool result = await _persistence.CreateBuildingAsync(newBuilding);

            // Assert
            Assert.True(result);
        }

        [Fact]
        public async void DeleteBuildingTestUnauthorized()
        {
            var deleteBuilding = new BuildingDTO
            {
                Id = BuildingData.First().Id
            };

            bool result = await _persistence.DeleteBuildingAsync(deleteBuilding); ;

            // Assert
            Assert.False(result);
        }

        [Fact]
        public async void DeleteBuildingTestAuthorized()
        {
            var deleteBuilding = new BuildingDTO
            {
               Id = BuildingData.First().Id
            };

            await _persistence.LoginAsync("tester", "Teszt123");
            bool result = await _persistence.DeleteBuildingAsync(deleteBuilding);

            // Assert
            Assert.True(result);
        }
    }
}
